'
' ORANGE Operating System
' FTPサーバー
' Copyright (C) 2020 PicoSoft
'
extern wifiStatus, 3
extern wifiLocalIP, 4
extern wifiServerNew, 7
extern wifiClientNew, 8
extern wifiServerBegin, 9
extern wifiServerHasClient, 10
extern wifiServerAvailable, 11
extern wifiClientConnected, 12
extern wifiClientAvailable, 13
extern wifiClientRead, 14
extern wifiClientStop, 15
extern wifiClientPrintln, 16
extern wifiClientPrint, 17
extern wifiClientWrite, 18
extern wifiClientReadBytes, 19
extern wifiClean, 23
extern yield, 100
extern fstatG, 103
const WL_CONNECTED = 3
const FALSE = 0
const TRUE = 1
const MAX_STRING = 128
const BEGIN = 0
const CONNECTION = 1
const IDLE = 2
const USER_IDENTITY = 3
const USER_PASSWORD = 4
const USER_REGISTRATION = 5
const USER_COMMAND = 6
const FTP_CMD_SIZE = 128
const FTP_CTRL_PORT = 21
const FTP_DATA_PORT_PASV = 50009
const MAX_PATH = 128
const FTP_BUF_SIZE = 1024
mstruct work {
	CMD_LINE rb FTP_CMD_SIZE
	COMMAND rb 5
	PARAMS rb FTP_CMD_SIZE
	CWD_NAME rb FTP_CMD_SIZE
	DATA_IP rb 16
	BUF rb MAX_PATH
	PATH rb MAX_PATH
	OLD rb MAX_PATH
}
mstruct stat {
	MODE rb 1
	ATIME rb 20
	MTIME rb 20
	CTIME rb 20
	SIZE rb 15
	INT32SIZE rb 4
}
mstruct dirInfo {
	BUF rb FTP_BUF_SIZE
	SORT.w rw 100
}
alias ftp.BUF, dirInfo.BUF
mstruct env {
	DRIVE rb 3 // "f":, "s:"
	PATH_F rb MAX_PATH // f:のカレントパス
	PATH_S rb MAX_PATH // s:のカレントパス
}
var transferStatus
var timer
var cmdp
var dataServer
var data
var dataPort
var file
var client

main()
end

def main() {
	var ftpServer
	var status
	var file

	if (wifiStatus() != WL_CONNECTED) {
		return
	}
	wifiClean()
	client = wifiClientNew()
	data = wifiClientNew()
	ftpServer = wifiServerNew(FTP_CTRL_PORT)
	wifiServerBegin(ftpServer)
	dataServer = wifiServerNew(FTP_DATA_PORT_PASV)
	wifiServerBegin(dataServer)
	status = BEGIN
	while TRUE
		if (inkey() == 0x1b) {
			print "ESC pressed."
			end
		}
		if (wifiServerHasClient(ftpServer)) {
			wifiClientStop(client)
			client = wifiServerAvailable(ftpServer)
		}
		if (status == BEGIN) {
			if (wifiClientConnected(client)) {
				disconnectClient()
			}
			status = CONNECTION
		} else if (status == CONNECTION) {
			abortTransfer()
			print "FTP Server waiting for connection"
			status = IDLE
		} else if (status == IDLE) {
			if (wifiClientConnected(client)) {
				print "Client connected!"
				wifiClientPrintln(client, mstrf("220--- Welcome to FTP for ORANGE-OS ---"))
				wifiClientPrintln(client, mstrf("220--- by PicoSoft ---"))
				wifiClientPrintln(client, mstrf("220 -- Ready"))
  				cmdp = 0
				status = USER_IDENTITY
			}
		} else if (0 < readLine()) {
			if (status == USER_IDENTITY) {
				if (userIdentity() == 0) {
					status = USER_PASSWORD
				} else {
					status = BEGIN
				}
			} else if (status == USER_PASSWORD) {
				if (userPassword() == 0) {
					status = USER_COMMAND
				} else {
					status = BEGIN
				}
			} else if (status == USER_COMMAND) {
				if (processCommand()) {
					print "ERROR"
					status = BEGIN
				}
			}
		} else if (!wifiClientConnected(client)) {
			status = CONNECTION
			print "client disconnected"
		}
		if (transferStatus == 1) {
			if (doRetrieve()) {
				transferStatus = 0
			}
		} else if (transferStatus == 2) {
			if (doStore()) {
				transferStatus = 0
			}
		}
	wend
}

def userIdentity() {
	if (mstrcmp(work.COMMAND, mstrf("USER")) != 0) {
		wifiClientPrintln(client, mstrf("500 Syntax error"))
	}
	if (mstrcmp(work.PARAMS, mstrf("pico")) != 0) {
		wifiClientPrintln(client, mstrf("530 user not found"))
	} else {
		wifiClientPrintln(client, mstrf("331 OK. Password required"))
		mstrcpy(work.CWD_NAME, mstrf("/"))
		return 0
	}
	pause 100
	return -1
}

def userPassword() {
	if (mstrcmp(work.COMMAND, mstrf("PASS")) != 0) {
		wifiClientPrintln(client, mstrf("500 Syntax error"))
	}
	if (mstrcmp(work.PARAMS, mstrf("pico")) != 0) {
		wifiClientPrintln(client, mstrf("530 "))
	} else {
		wifiClientPrintln(client, mstrf("230 OK."))
		return 0
	}
	pause 100
	return -1
}

' クライアントからのコマンドを処理します。
def processCommand() {
	var p
	var i
	if (mstrcmp(work.COMMAND, mstrf("FEAT")) == 0) {
		wifiClientPrintln(client, mstrf("211-Extensions suported:"))
		wifiClientPrintln(client, mstrf(" MLSD"))
		wifiClientPrintln(client, mstrf("211 End."))
	} else if (mstrcmp(work.COMMAND, mstrf("PWD")) == 0) {
		wifiClientPrint(client, mstrf("257 \""))
		wifiClientPrint(client, work.CWD_NAME)
		wifiClientPrintln(client, mstrf("\" is your current directory"))
	} else if (mstrcmp(work.COMMAND, mstrf("CWD")) == 0) {
		if (work.PARAMS[0] == '.') {
			wifiClientPrint(client, mstrf("257 \""))
			wifiClientPrint(client, work.CWD_NAME)
			wifiClientPrintln(client, mstrf("\" is your current directory"))
		} else {
			wifiClientPrint(client, mstrf("250 Ok. Current directory is "))
			wifiClientPrintln(client, work.CWD_NAME)
		}
	} else if (mstrcmp(work.COMMAND, mstrf("TYPE")) == 0) {
		wifiClientPrint(client, mstrf("200 TYPE is now "))
		if (work.PARAMS[0] == 'A') {
			wifiClientPrintln(client, mstrf("ASII"))
		} else if (work.PARAMS[0] == 'I') {
			wifiClientPrintln(client, mstrf("8-bit binary"))
		} else {
			wifiClientPrintln(client, mstrf("504 Unknown TYPE"))
		}
	} else if (mstrcmp(work.COMMAND, mstrf("PASV")) == 0) {
		if (wifiClientConnected(data)) {
			wifiClientStop(data)
		}
		wifiLocalIP(work.DATA_IP)
		dataPort = FTP_DATA_PORT_PASV
		p = work.DATA_IP
		i = 0
		while p[i] != 0
			if (p[i] == '.') {
				p[i] = ','
			}
			i = i + 1
		wend
		wifiClientPrintln(client, mstrf("227 Entering Passive Mode (%s,%d,%d).", work.DATA_IP, dataPort >> 8, dataPort & 255))
	} else if (mstrcmp(work.COMMAND, mstrf("LIST")) == 0) {
		listCommand(FALSE)
	} else if (mstrcmp(work.COMMAND, mstrf("MLSD")) == 0) {
		listCommand(TRUE)
	} else if (mstrcmp(work.COMMAND, mstrf("RETR")) == 0) {
		if (work.PARAMS[0] == 0) {
			wifiClientPrintln(client, mstrf("501 No file name"))
		} else {
			getFullPath(work.PATH, work.PARAMS)
			file = fopen(work.PATH, 0)
			if (file == -1) {
				wifiClientPrintln(client, mstrf("550 File %s not found", work.PARAMS))
			} else if (!dataConnect()) {
				wifiClientPrintln(client, mstrf("425 No data connection"))
			} else {
				wifiClientPrintln(client, mstrf("150-Connected to port %u", dataPort))
				fstat(work.PATH, stat.MODE)
				trim(work.BUF, stat.SIZE)
				wifiClientPrintln(client, mstrf("150 %s bytes to download", work.BUF))
				transferStatus = 1
			}
		}
	} else if (mstrcmp(work.COMMAND, mstrf("STOR")) == 0) {
		if (work.PARAMS[0] == 0) {
			wifiClientPrintln(client, mstrf("501 No file name"))
		} else {
			getFullPath(work.PATH, work.PARAMS)
			file = fopen(work.PATH, 1)
			if (file == -1) {
				wifiClientPrintln(client, mstrf("451 Can't open/create %s", work.PARAMS))
			} else if (!dataConnect()) {
				wifiClientPrintln(client, mstrf("425 No data connection"))
				fclose(file)
			} else {
				wifiClientPrintln(client, mstrf("150 Connected to port %u", dataPort))
				transferStatus = 2
			}
		}
	} else if (mstrcmp(work.COMMAND, mstrf("DELE")) == 0) {
		if (work.PARAMS[0] == 0) {
			wifiClientPrintln(client, mstrf("501 No file name"))
		} else {
			getFullPath(work.PATH, work.PARAMS)
			if (fstat(work.PATH, stat.MODE) < 0) {
				wifiClientPrintln(client, mstrf("550 No file %s not found", work.PARAMS))
			} else {
				if (remove(work.PATH) == 0) {
					wifiClientPrintln(client, mstrf("250 Deleted %s", work.PARAMS))
				} else {
					wifiClientPrintln(client, mstrf("450 Can't delete %s", work.PARAMS))
				}
			}
		}
	} else if (mstrcmp(work.COMMAND, mstrf("RNFR")) == 0) {
		work.OLD[0] = 0
		if (work.PARAMS[0] == 0) {
			wifiClientPrintln(client, mstrf("501 No file name"))
		} else {
			getFullPath(work.OLD, work.PARAMS)
			if (fstat(work.OLD, stat.MODE) < 0) {
				wifiClientPrintln(client, mstrf("550 No file %s not found", work.PARAMS))
			} else {
				wifiClientPrintln(client, mstrf("350 RNFR accepted - file exists, ready for destination"))
			}
		}
	} else if (mstrcmp(work.COMMAND, mstrf("RNTO")) == 0) {
		if (work.OLD[0] == 0) {
			wifiClientPrintln(client, mstrf("503 Need RNFR before RNTO"))
		} else if (work.PARAMS[0] == 0) {
			wifiClientPrintln(client, mstrf("501 No file name"))
		} else {
			getFullPath(work.PATH, work.PARAMS)
			if (0 <= fstat(work.PATH, stat.MODE)) {
				wifiClientPrintln(client, mstrf("553 %s already exists", work.PARAMS))
			} else {
				if (rename(work.OLD, work.PATH) == 0) {
					wifiClientPrintln(client, mstrf("250 File successfully renamed or moved"))
				} else {
					wifiClientPrintln(client, mstrf("451 Rename/move failure"))
				}
			}
		}
	} else if (mstrcmp(work.COMMAND, mstrf("QUIT")) == 0) {
		disconnectClient()
		return -1
	} else {
		wifiClientPrintln(client, mstrf("500 Unknown command"))
	}
	return 0
}

def listCommand(isMlsd) {
	var count
	var i
	if (!dataConnect()) {
		wifiClientPrintln(client, mstrf("425 No data connection"))
	} else {
		wifiClientPrintln(client, mstrf("150 Accepted data connection"))
		count = dir(work.CWD_NAME, dirInfo.BUF)
		makePointerArray(count, dirInfo.SORT.w, dirInfo.BUF)
		for i = 0 to count - 1
			printFileInfo(dirInfo.SORT.w[i], isMlsd)
		next
		if (isMlsd) {
			wifiClientPrintln(client, mstrf("226-options: -a -l"))
		}
		wifiClientPrintln(client, mstrf("226 %d matches total", count))
		wifiClientStop(data)
	}
}


def printFileInfo(name, isMlsd) {
	var h
	getFullPath(work.PATH, name)
	fstatG(work.PATH, stat.MODE, 1) // GMT取得用fstat
	if (isMlsd) { // MLSD
		if (stat.MODE[0] = 1) {
			wifiClientPrint(data, mstrf("Type=dir;Size="))
		} else {
			wifiClientPrint(data, mstrf("Type=file;Size="))
		}
		trim(work.BUF, stat.SIZE)
		wifiClientPrint(data, mstrf("%s;modify=", work.BUF))
		trim(work.BUF, stat.MTIME)
		wifiClientPrintln(data, mstrf("%s; %s", work.BUF, name))
	} else { // LIST
		work.BUF[0] = stat.MTIME[5]
		work.BUF[1] = stat.MTIME[6]
		work.BUF[2] = '-'
		work.BUF[3] = stat.MTIME[8]
		work.BUF[4] = stat.MTIME[9]
		work.BUF[5] = '-'
		work.BUF[6] = stat.MTIME[0]
		work.BUF[7] = stat.MTIME[1]
		work.BUF[8] = stat.MTIME[2]
		work.BUF[9] = stat.MTIME[3]
		work.BUF[10] = ' '
		work.BUF[11] = ' '
		work.BUF[12] = 0
		wifiClientPrint(data, work.BUF)
		h = (stat.MTIME[11] - '0') * 10 + (stat.MTIME[12] - '0')
		if (h < 12) {
			wifiClientPrint(data, mstrf("%d:", h))
		} else {
			wifiClientPrint(data, mstrf("%d:", h - 12))
		}
		work.BUF[0] = stat.MTIME[14]
		work.BUF[1] = stat.MTIME[15]
		work.BUF[2] = 0
		wifiClientPrint(data, work.BUF)
		if (h < 12) {
			wifiClientPrint(data, mstrf("AM "))
		} else {
			wifiClientPrint(data, mstrf("PM "))
		}
		if (stat.MODE[0] = 1) {
			wifiClientPrintln(data, mstrf("<DIR> %s", name))
		} else {
			trim(work.BUF, stat.SIZE)
			wifiClientPrintln(data, mstrf("%s %s", work.BUF, name))
		}
	}
}

def doRetrieve() {
	var size
	size = fread(file, ftp.BUF, FTP_BUF_SIZE)
	if (0 < size) {
		wifiClientWrite(data, ftp.BUF, size)
		return 0
	}
	closeTransfer()
	return -1
}

def doStore() {
	var size
	size = wifiClientReadBytes(data, ftp.BUF, FTP_BUF_SIZE)
	if (0 < size) {
		fwrite(file, ftp.BUF, size)
		return 0
	}
	if (wifiClientConnected(data)) {
		return 0
	} else {
		closeTransfer()
		return -1
	}
}

def closeTransfer() {
	wifiClientPrintln(client, mstrf("226 File successfully transferred"))
	fclose(file)
	wifiClientStop(data)
}

def trim(d, s) {
	var c
	while TRUE
		c = s[0]
		if (c == 0) {
			break
		}
		if (c != ' ' && c != ',' && c != '/' && c != ':') {
			d[0] = s[0]
			d = d + 1
		}
		s = s + 1
	wend
	d[0] = 0
}

' クライアントからのリクエストを読んで、コマンドとパラメーターに分解します。
def readLine() {
	var rc
	var c
	var p
	var len
	rc = -1
	if (wifiClientAvailable(client)) {
		c = wifiClientRead(client)
		if (c == '\\') {
			c = '/'
		}
		if (c != '\r') {
			if (c == '\n') {
				work.CMD_LINE[cmdp] = 0
				mprint work.CMD_LINE: print
				work.COMMAND[0] = 0
				work.PARAMS[0] = 0
				if (cmdp == 0) { // 空行
					rc = 0
				} else {
					rc = cmdp // 文字数
					p = mstrchr(work.CMD_LINE, ' ')
					if (p != -1) {
						len = p - work.CMD_LINE
						mmove(work.COMMAND, work.CMD_LINE, len)
						work.COMMAND[len] = 0
						mstrcpy(work.PARAMS, p + 1)
					} else {
						mstrcpy(work.COMMAND, work.CMD_LINE)
					}
					cmdp = 0
					if (4 < mstrlen(work.COMMAND)) {
						rc = -2
					}
				}
			} else {
				if (cmdp < FTP_CMD_SIZE) {
					work.CMD_LINE[cmdp] = c
					cmdp = cmdp + 1
			 	} else {
					rc = -2
				}
			}
		}
	}
	if (rc == -2) {
		cmdp = 0
		wifiClientPrintln(client, mstrf("500 Syntax error"))
	}
	return rc
}

def dataConnect() {
	if (!wifiClientConnected(data)) {
		while (!wifiServerHasClient(dataServer))
			yield()
		wend
		if (wifiServerHasClient(dataServer)) {
			wifiClientStop(data)
			data = wifiServerAvailable(dataServer)
		}
	}
	return wifiClientConnected(data)
}

def disconnectClient() {
	print "Disconnecting client"
	abortTransfer()
	wifiClientPrintln(client, mstrf("221 Goodbye"))
	wifiClientStop(client)
}

def abortTransfer(file) {
	if (0 < transferStatus) {
		fclose(file)
		wifiClientStop(data)
		wifiClientPrintln(client, mstrf("426 Transfer aborted"))
		print "Transfer aborted!"
	}
	transferStatus = 0
}

' ポインター配列を作ります。
def makePointerArray(count, d.w, s) {
	var i
	for i = 0 to count - 1
		d.w[i] = s
		s = s + mstrlen(s) + 1
	next
}

' 完全パス名を作ります。
def getFullPath(fullPath, name) {
	var p
	var currentPath
	getenv "DRIVE", env.DRIVE // カレントドライブ取得
	getenv "PATH_F", env.PATH_F // カレントディレクトリ取得
	getenv "PATH_S", env.PATH_S // カレントディレクトリ取得
	p = name
	if (p[1] == ':') { // ドライブ指定あり
		fullPath[0] = p[0]
		fullPath[1] = p[1]
		fullPath[2] = 0
		p = p + 2
	} else {
		mstrcpy(fullPath, env.DRIVE) // カレントドライブ
	}
	if (p[0] == '/') { // ルートからの指定
		mstrcat(fullPath, p)
		return
	}
	if (fullPath[0] == 'f') {
		currentPath = env.PATH_F
	} else {
		currentPath = env.PATH_S
	}
	mstrcat(fullPath, currentPath)
	if (mstrcmp(currentPath, mstrf("/")) != 0) {
		mstrcat(fullPath, mstrf("/"))
	}
	mstrcat(fullPath, p)
}
